#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
"""This module runs a Jupyter Python notebook with nbconvert and print the outputs.
This is a driver script auto-generated by Oracle ADS.

The following environment variables are used:
JOB_RUN_NOTEBOOK:
    The relative path of the jupyter Python notebook to be executed.
NOTEBOOK_EXCLUDE_TAGS:
    Optional, a list of tags serialized to JSON string.
    Notebook cells with one of the tags will be excluded from running.
NOTEBOOK_ENCODING:
    Optional, the encoding for opening the notebook.
OUTPUT_URI:
    Optional, object storage URI for saving files from the output directory.
"""
import logging
import json
import os
import subprocess
from typing import Optional
from urllib.parse import urlparse

import nbformat
import oci
from nbconvert.preprocessors import ExecutePreprocessor


logger = logging.getLogger(__name__)


class ADSExecutePreprocessor(ExecutePreprocessor):
    """Customized Execute Preprocessor for running notebook."""

    def __init__(self, exclude_tags=None, **kw):
        """Initialize the preprocessor

        Parameters
        ----------
        exclude_tags : list, optional
            A list of cell tags, notebook cells with any of these cell tag will be skipped.
            Defaults to None.
        """
        self.exclude_tags = exclude_tags
        super().__init__(**kw)

    @staticmethod
    def _print_cell_outputs(cell):
        """Prints the outputs of a notebook cell."""
        for output in cell.outputs:
            output_type = output.get("output_type")
            if output_type == "stream":
                # stream outputs includes line break already
                print(output.text, end="")
            elif output_type == "execute_result":
                # execute_result may contain text/plain
                text = output.get("data", {}).get("text/plain", [])
                # The value could be str or list of str
                if isinstance(text, list):
                    for line in text:
                        print(line)
                else:
                    print(text)

    def preprocess_cell(self, cell, resources, *args, **kwargs):
        """Runs the notebook cell and print out the outputs"""
        # Skip the cell if any of the cell tags matching an exclude tag.
        if self.exclude_tags:
            # Log an error message if there is an error reading the cell tags,
            # and continue to run the cell.
            try:
                cell_tags = cell.get("metadata", {}).get("tags", [])
                for tag in cell_tags:
                    if tag in self.exclude_tags:
                        return cell, resources
            except Exception as ex:
                logger.error("An error occurred when reading cell tags: %s", str(ex))
        # Run the cell
        cell, resources = super().preprocess_cell(cell, resources, *args, **kwargs)
        # Print cell output
        if hasattr(cell, "outputs"):
            # Log a message if there is an error getting the cell output,
            # and continue to run the next cell.
            try:
                self._print_cell_outputs(cell)
            except Exception as ex:
                logger.error("An error occurred when reading cell outputs: %s", str(ex))
        return cell, resources


class OCIHelper:
    @staticmethod
    def init_oci_client(client_class):
        """Initializes OCI client with API key or Resource Principal.

        Parameters
        ----------
        client_class :
            The class of OCI client to be initialized.
        """
        if os.environ.get("OCI_RESOURCE_PRINCIPAL_VERSION"):
            logger.info(
                "Initializing %s with Resource Principal...", client_class.__name__
            )
            client = client_class(
                {}, signer=oci.auth.signers.get_resource_principals_signer()
            )
        else:
            logger.info("Initializing %s with API Key...", {client_class.__name__})
            client = client_class(oci.config.from_file())
        return client

    @staticmethod
    def copy_outputs(output_dir: str, output_uri: dict) -> None:
        """Copies the output files to object storage bucket.

        Parameters
        ----------
        output_dir : str
            Path of the output directory containing files to be copied.
        output_uri : str
            URI of the object storage path to store the output files.
        """
        output_dir = os.path.abspath(os.path.expanduser(output_dir))
        if not os.path.exists(output_dir):
            logger.error("Output directory %s not found.", output_dir)
            return
        if not output_uri:
            logger.error(
                "OUTPUT_URI is not defined in environment variable. No file is copied."
            )
            return
        logger.info("Copying files in %s to %s...", output_dir, output_uri)
        parsed = urlparse(output_uri)
        bucket_name = parsed.username
        namespace = parsed.hostname
        prefix = parsed.path
        oci_os_client = OCIHelper.init_oci_client(
            oci.object_storage.ObjectStorageClient
        )

        if not prefix:
            prefix = ""
        prefix = prefix.strip("/")

        for path, _, files in os.walk(output_dir):
            for name in files:
                file_path = os.path.join(path, name)

                with open(file_path, "rb") as pkf:
                    # Get the relative path of the file to keep the directory structure
                    relative_path = os.path.relpath(file_path, output_dir)
                    if prefix:
                        file_prefix = os.path.join(prefix, relative_path)
                    else:
                        # Save file to bucket root if prefix is empty.
                        file_prefix = relative_path

                    logger.debug(
                        f"Saving {relative_path} to {bucket_name}@{namespace}/{file_prefix}"
                    )

                    oci_os_client.put_object(
                        namespace,
                        bucket_name,
                        file_prefix,
                        pkf,
                    )


def run_notebook(
    notebook_path: str,
    working_dir: Optional[str] = None,
    exclude_tags: Optional[list] = None,
) -> None:
    """Runs a notebook

    Parameters
    ----------
    notebook_path : str
        The path of the notebook
    working_dir : str, optional
        The working directory for running the notebook, by default None.
        If this is None, the same directory of the notebook_path will be used.
    exclude_tags : list, optional
        Tags for excluding cells, by default None
    """
    # Read the notebook
    encoding = os.environ.get("NOTEBOOK_ENCODING", "utf-8")
    with open(notebook_path, encoding=encoding) as f:
        nb = nbformat.read(f, as_version=4)

    # Working/Output directory
    if not working_dir:
        working_dir = os.path.dirname(notebook_path)

    # The path of the output notebook with results/plots
    notebook_filename_out = os.path.join(working_dir, os.path.basename(notebook_path))

    ep = ADSExecutePreprocessor(exclude_tags=exclude_tags, kernel_name="python")

    from nbconvert.preprocessors import CellExecutionError

    try:
        ep.preprocess(nb, {"metadata": {"path": working_dir}})
    except CellExecutionError:
        msg = "Error executing the notebook.\n\n"
        msg += f'See notebook "{notebook_filename_out}" for the traceback.'
        logger.error(msg)
        raise
    finally:
        with open(notebook_filename_out, mode="w", encoding=encoding) as f:
            nbformat.write(nb, f)


def run_driver() -> None:
    """Runs the driver to execute a notebook."""
    try:
        # Ignore the error as conda-unpack may not exists
        subprocess.check_output("conda-unpack", stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as ex:
        logger.debug("conda-unpack exits with non-zero return code %s", ex.returncode)
        logger.debug(ex.output)
    notebook_file_path = os.path.join(
        os.path.dirname(__file__), os.environ.get("JOB_RUN_NOTEBOOK")
    )
    output_dir = os.path.join(os.path.dirname(__file__), "outputs")
    # Create the output directory
    os.makedirs(output_dir, exist_ok=True)
    # Exclude tags
    tags = os.environ.get("NOTEBOOK_EXCLUDE_TAGS")
    if tags:
        tags = json.loads(tags)
        logger.info("Excluding cells with any of the following tags: %s", tags)
    # Run the notebook
    run_notebook(notebook_file_path, working_dir=output_dir, exclude_tags=tags)
    # Save the outputs
    OCIHelper.copy_outputs(output_dir, os.environ.get("OUTPUT_URI"))


if __name__ == "__main__":
    run_driver()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
"""This is a driver script auto-generated by Oracle ADS to run Python script in OCI Data Science Jobs.

The following environment variables are used:
CODE_ENTRYPOINT:
    Relative path to the entry script/module file in the user code directory.
ENTRY_FUNCTION:
    Optional, function name in the entry script/module to be invoked.
    If this is not specified, the entry script will be run as Python script.
PYTHON_PATH:
    Optional, additional paths to be added to sys.path for looking up modules and packages.
    The root of the user code will be added by default.
    Multiple paths can be separated by os.pathsep, which is colon(:) for Linux and Mac, semicolon(;) for Windows.
OUTPUT_DIR:
    Optional, output directory to be copied to object storage.
OUTPUT_URI:
    Optional, object storage URI for saving files from the output directory.
WORKING_DIR:
    Optional, the working directory of the user code.
    This can be specified as a relative path from /home/datascience/decompressed_artifact/code
    If this is not specified, the working directory will be /home/datascience/decompressed_artifact/code

This module requires the following package:
oci


"""
import importlib
import json
import logging
import os
import runpy
import subprocess
import sys
from typing import Optional
from urllib.parse import urlparse

import oci


logger = logging.getLogger(__name__)


# Constants
CONST_METADATA = "METADATA"
USER_CODE_DIR = "code"


class OCIHelper:
    @staticmethod
    def init_oci_client(client_class):
        """Initializes OCI client with API key or Resource Principal.

        Parameters
        ----------
        client_class :
            The class of OCI client to be initialized.
        """
        if os.environ.get("OCI_RESOURCE_PRINCIPAL_VERSION"):
            logger.info(
                "Initializing %s with Resource Principal...", client_class.__name__
            )
            client = client_class(
                {}, signer=oci.auth.signers.get_resource_principals_signer()
            )
        else:
            logger.info("Initializing %s with API Key...", {client_class.__name__})
            client = client_class(oci.config.from_file())
        return client

    @staticmethod
    def copy_outputs(output_dir: str, output_uri: dict) -> None:
        """Copies the output files to object storage bucket.

        Parameters
        ----------
        output_dir : str
            Path of the output directory containing files to be copied.
        output_uri : str
            URI of the object storage path to store the output files.
        """
        output_dir = os.path.abspath(os.path.expanduser(output_dir))
        if not os.path.exists(output_dir):
            logger.error("Output directory %s not found.", output_dir)
            return
        if not output_uri:
            logger.error(
                "OUTPUT_URI is not defined in environment variable. No file is copied."
            )
            return
        logger.debug("Copying files in %s to %s...", output_dir, output_uri)
        parsed = urlparse(output_uri)
        bucket_name = parsed.username
        namespace = parsed.hostname
        prefix = parsed.path
        oci_os_client = OCIHelper.init_oci_client(
            oci.object_storage.ObjectStorageClient
        )

        if not prefix:
            prefix = ""
        prefix = prefix.strip("/")

        for path, _, files in os.walk(output_dir):
            for name in files:
                file_path = os.path.join(path, name)

                with open(file_path, "rb") as pkf:
                    # Get the relative path of the file to keep the directory structure
                    relative_path = os.path.relpath(file_path, output_dir)
                    if prefix:
                        file_prefix = os.path.join(prefix, relative_path)
                    else:
                        # Save file to bucket root if prefix is empty.
                        file_prefix = relative_path

                    logger.debug(
                        f"Saving {relative_path} to {bucket_name}@{namespace}/{file_prefix}"
                    )

                    oci_os_client.put_object(
                        namespace,
                        bucket_name,
                        file_prefix,
                        pkf,
                    )


class JobRunner:
    """Contains methods for running the job."""

    def __init__(self, code_dir: str) -> None:
        """Initialize the job runner

        Parameters
        ----------
        code_dir : str
            the path to the directory containing the user code.
        """
        self.code_dir = code_dir

    def setup_python_path(self, python_paths: str):
        """Adds additional python paths.

        Parameters
        ----------
        python_paths : str
            Additional python paths to be added to sys.path.
            Multiple paths can be separated by os.pathsep, which is colon(:) for Linux and Mac.

        """
        python_paths = python_paths.split(os.pathsep)
        python_paths.append(self.code_dir)
        for path in python_paths:
            python_path = os.path.abspath(os.path.expanduser(path))
            if python_path not in sys.path:
                sys.path.append(python_path)
        logger.debug("Python Path: %s", sys.path)
        return self

    def _run_function(self, module_path: str, entry_function: str, argv: list):
        """Runs the entry function in module specified by module path.

        Parameters
        ----------
        module_path : str
            The path to the module containing the entry function.
        entry_function : str
            The name of the entry function.
        argv : list
            Argument list from command line.
        This list will be parsed into positional arguments and keyword arguments.
        """
        spec = importlib.util.spec_from_file_location("module", module_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        method = getattr(module, entry_function)

        args, kwargs = ArgumentParser(argv).parse()

        logger.debug(
            "Invoking method: %s with args: %s, kwargs: %s",
            entry_function,
            args,
            kwargs,
        )
        method(*args, **kwargs)

    def run(
        self,
        entrypoint: str,
        entry_function: Optional[str] = None,
        argv: Optional[list] = None,
    ):
        """Runs the user code

        Parameters
        ----------
        entrypoint : str
            Path to the file serve as the entrypoint.
        entry_function : str, optional
            Name of the function in the entrypoint, by default None.
            If this is not specified, the entrypoint will be run as a python script.
        argv : list, optional
            Arguments to be parsed and passed into the entry_function, by default None.
            The arguments here are used only when running the entry_function.
            When entry_function is None, sys.argv[1:] will be used as command line argument.

        """
        if not argv:
            argv = []

        module_path = entrypoint

        if entry_function:
            logger.info(f"Running function: {entry_function} in {module_path}...")
            self._run_function(module_path, entry_function, argv)
        else:
            logger.info(f"Running script: {module_path}")
            runpy.run_path(module_path, run_name="__main__")
        return self


class ArgumentParser:
    """Contains methods for parsing arguments for entry function."""

    def __init__(self, argument_list: list) -> None:
        """Initialize the parser with a list of arguments

        Parameters
        ----------
        argument_list : list
            A list of arguments.
        """
        self.argument_list = argument_list

    @staticmethod
    def decode_arg(val):
        """Decodes the value of the argument if it is a JSON payload.

        Parameters
        ----------
        val : str
            The argument value in a string.

        Returns
        -------
        Any
            None, if the val is None.
            String value, if the val is a string but not a JSON payload.
            Otherwise, the object after JSON decoded.
        """
        if val is None:
            return None
        try:
            return json.loads(val)
        except json.decoder.JSONDecodeError:
            return val

    @staticmethod
    def join_values(value_list):
        """Joins the values of a keyword argument.

        Parameters
        ----------
        value_list : list
            Values in a list of strings.

        Returns
        -------
        str or None
            The value of the argument as a string.
        """
        if value_list:
            return " ".join(value_list)
        return None

    def parse(self):
        """Parses the arguments

        Returns
        -------
        (list, dict)
            A tuple of positional arguments (list) and keyword arguments (dict).
        """
        args = []
        kwargs = {}
        parsing_kwargs = False
        key = None
        val = []
        for arg in self.argument_list:
            arg = str(arg)
            if len(arg) > 2 and arg.startswith("--"):
                if key:
                    # Save previous key and val
                    kwargs[key] = self.join_values(val)
                parsing_kwargs = True
                key = arg[2:]
                # Reset val
                val = []
            elif parsing_kwargs:
                val.append(arg)
            else:
                args.append(arg)
        # Save the last key and val
        if key:
            kwargs[key] = self.join_values(val)

        args = [self.decode_arg(arg) for arg in args]
        kwargs = {k: self.decode_arg(v) for k, v in kwargs.items()}
        return args, kwargs


def main():
    """The main function for running the job."""
    job_run_ocid = os.environ.get("JOB_RUN_OCID")
    logger.info("Job Run ID is: %s", job_run_ocid)

    # Conda unpack
    try:
        # Ignore the error as conda-unpack may not exists
        subprocess.check_output("conda-unpack", stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as ex:
        logger.debug("conda-unpack exits with non-zero return code %s", ex.returncode)
        logger.debug(ex.output)

    code_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), USER_CODE_DIR)
    # Change the working diretory to the root of user code
    if os.environ.get("WORKING_DIR"):
        working_dir = os.path.join(code_dir, os.environ.get("WORKING_DIR"))
    else:
        working_dir = code_dir
    os.chdir(working_dir)
    # Add working dir to sys.path
    if working_dir not in sys.path:
        sys.path.append(working_dir)

    # Run user code
    runner = JobRunner(code_dir)
    runner.setup_python_path(os.environ.get("PYTHON_PATH", "")).run(
        entrypoint=os.environ.get("CODE_ENTRYPOINT"),
        entry_function=os.environ.get("ENTRY_FUNCTION"),
        argv=sys.argv[1:],
    )

    logger.info("Job completed.")

    # Copy outputs
    if "OUTPUT_DIR" in os.environ:
        logger.debug("OUTPUT_DIR is: %s", os.environ["OUTPUT_DIR"])
        OCIHelper.copy_outputs(
            output_dir=os.environ["OUTPUT_DIR"],
            output_uri=os.environ.get("OUTPUT_URI"),
        )
    else:
        logger.debug("OUTPUT_DIR is not configured. Skipping copy artifacts")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import json
import logging
import os
from functools import lru_cache
from typing import Dict, Tuple

import requests
import yaml
from ads.common.utils import PAR_LINK
from ads.common.object_storage_details import ObjectStorageDetails
from cerberus import DocumentError, Validator

MODEL_PROVENANCE_SCHEMA_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "schemas",
    "model_provenance_schema.yaml",
)
INFERENCE_ENV_SCHEMA_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "schemas",
    "inference_env_info_schema.yaml",
)
TRAINING_ENV_SCHEMA_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "schemas",
    "training_env_info_schema.yaml",
)


class SchemaValidator:
    """
    Base Schema Validator which validate yaml file.
    """

    def __init__(self, schema_file_path: str) -> "SchemaValidator":
        """Initiate a SchemaValidator instance.

        Parameters
        ----------
        schema_file_path: (str)
            schema file path. The schema is used to validate the yaml file.

        Returns
        -------
        SchemaValidator
            A SchemaValidator instance.
        """
        self.schema_file_path = schema_file_path
        self._schema_validator = self._load_schema_validator()

    def validate(self, document: Dict) -> bool:
        """Validate the schema.

        Parameters
        ----------
        document: (Dict)
            yaml file content to validate.

        Raises
        ------
        DocumentError: Raised when the validation schema is missing, has the wrong format or contains errors.

        Returns
        -------
        bool
            validation result.
        """
        v = Validator()
        res = v.validate(document, self._schema_validator)
        if not res:
            raise DocumentError(v.errors)
        return res

    def _load_schema_validator(self):
        """load the schema validator to validate the schema.

        Returns
        -------
        dict
            Schema validator.
        """
        with open(self.schema_file_path, encoding="utf-8") as schema_file:
            ext = os.path.splitext(self.schema_file_path)[-1].lower()
            if ext in [".yaml", ".yml"]:
                schema_validator = yaml.load(schema_file, yaml.FullLoader)
            elif ext in [".json"]:
                schema_validator = json.load(schema_file)
            else:
                raise NotImplementedError(f"{ext} format schema is not supported.")
        return schema_validator


@lru_cache(maxsize=1)
def get_service_packs(namespace: str, bucketname: str) -> Tuple[Dict, Dict]:
    """Get the service pack path mapping and service pack slug mapping.
    Note: deprecated packs are also included.

    Parameters
    ----------
    namespace: str
        namespace of the service pack.
    bucketname: str
        bucketname of the service pack.

    Returns
    -------
    (Dict, Dict)
        Service pack path mapping(service pack path -> (slug, python version))
        and the service pack slug mapping(service pack slug -> (pack path, python version)).
    """
    service_pack_path_mapping = {}
    service_pack_slug_mapping = {}
    try:
        response = requests.request("GET", PAR_LINK)
        if response.ok:
            service_pack_list = response.json().get("service_packs")
            for service_pack in service_pack_list:
                # Here we need to replace the namespace and bucketname
                # with the bucket and namespace of the region that
                # user is in. The reason is that the mapping is generated
                # from the index.json file which has static namespace
                # and bucket of prod. however, namespace will change based
                # on the region. also, dev has different bucketname.
                pack_path = ObjectStorageDetails(
                    bucket=bucketname,
                    namespace=namespace,
                    filepath=ObjectStorageDetails.from_path(
                        service_pack.get("pack_path")
                    ).filepath,
                ).path
                service_pack_path_mapping[pack_path] = (
                    service_pack.get("slug"),
                    service_pack.get("python"),
                )
                service_pack_slug_mapping[service_pack["slug"]] = (
                    pack_path,
                    service_pack.get("python"),
                )
    except Exception as e:
        logging.warning(e)
        logging.warning(
            "Failed to auto generate the `service_pack_path_mapping`. "
            "Hence, the slug, environment type and python version field are not populated."
        )
    return service_pack_path_mapping, service_pack_slug_mapping
